#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <GL/glut.h>
#include "sdr.h"

#ifdef FREEGLUT
#include <GL/freeglut_ext.h>
#define MOUSEWHEEL
#endif

void redraw(void);
void set_material_color(float r, float g, float b, float a);
void setup_lights(void);
void key_handler(unsigned char key, int x, int y);
void key_up_handler(unsigned char key, int x, int y);
void skey_handler(int key, int x, int y);
void button_handler(int bn, int state, int x, int y);
void mouse_handler(int x, int y);
void reshape(int x, int y);
#ifdef MOUSEWHEEL
void wheel_handler(int unk, int dir, int x, int y);
#endif
void menu_handler(int val);
void init_sdr(void);

int xres, yres;

float cam_x, cam_y = 0.6, cam_z = 4;
float cam_rot, cam_pitch = 35.0;

float pan_offs = 0.025;

unsigned int prog;

float roughness = 0.24;
float specularity = 0.86;
float ior = 3.2;
float spec_pow = 60.0;

enum {DSDR_LAMBERT, DSDR_OREN_NAYAR};
enum {SSDR_PHONG, SSDR_BLINN, SSDR_COOK_TORR};

int dif_sdr = DSDR_OREN_NAYAR;
int spec_sdr = SSDR_COOK_TORR;

int main(int argc, char **argv)
{
	unsigned int vs, ps, pslib;
	int dif_menu, spec_menu, i;

	for(i=1; i<argc; i++) {
		if(argv[i][0] == '-' && argv[i][2] == 0) {
			switch(argv[i][1]) {
			case 'd':
				if(strcmp(argv[++i], "lambert") == 0) {
					dif_sdr = DSDR_LAMBERT;
				} else if(strcmp(argv[i], "oren-nayar") == 0) {
					dif_sdr = DSDR_OREN_NAYAR;
				} else {
					printf("-d (diffuse model) choices:\n");
					printf("  lambert        \"Photometria sive de mensura de gratibus luminis, colorum et umbrae\", 1760\n");
					printf("  oren-nayar     \"Generalization of Lambert's Reflectance Model\", SIGGRAPH 1994\n");
					return EXIT_FAILURE;
				}
				break;

			case 's':
				if(strcmp(argv[++i], "phong") == 0) {
					spec_sdr = SSDR_PHONG;
				} else if(strcmp(argv[i], "cook-torrance") == 0) {
					spec_sdr = SSDR_COOK_TORR;
				} else {
					printf("-s (specular model) choices:\n");
					printf("  phong          \"Illumination for Computer Generated Pictures\", CACM 1975\n");
					printf("  blinn          \"Models of Light Reflection for Computer Synthesized Pictures\", SIGGRAPH 1977\n");
					printf("  cook-torrance  \"A Reflectance Model for Computer Graphics\", SIGGRAPH 1981\n");
					return EXIT_FAILURE;
				}
				break;

			default:
				fprintf(stderr, "unrecognized argument: %s\n", argv[i]);
				return EXIT_FAILURE;
			}
		}
	}

	glutInitWindowSize(800, 600);
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DEPTH | GLUT_DOUBLE);
	glutCreateWindow("1.roughness  2.specularity  3.ior  4.phong power");
	xres = glutGet(GLUT_WINDOW_WIDTH);
	yres = glutGet(GLUT_WINDOW_HEIGHT);

	glutKeyboardFunc(key_handler);
	glutKeyboardUpFunc(key_up_handler);
	glutSpecialFunc(skey_handler);
	glutMotionFunc(mouse_handler);
	glutMouseFunc(button_handler);
	glutReshapeFunc(reshape);
	glutDisplayFunc(redraw);
#ifdef MOUSEWHEEL
	glutMouseWheelFunc(wheel_handler);
#endif

	/* create the menus */
	dif_menu = glutCreateMenu(menu_handler);
	glutAddMenuEntry("lambert", DSDR_LAMBERT);
	glutAddMenuEntry("oren-nayar", DSDR_OREN_NAYAR);

	spec_menu = glutCreateMenu(menu_handler);
	glutAddMenuEntry("phong", SSDR_PHONG + 10);
	glutAddMenuEntry("blinn", SSDR_BLINN + 10);
	glutAddMenuEntry("cook-torrance", SSDR_COOK_TORR + 10);

	glutCreateMenu(menu_handler);
	glutAddSubMenu("diffuse models", dif_menu);
	glutAddSubMenu("specular models", spec_menu);

	glutAttachMenu(GLUT_RIGHT_BUTTON);


	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);

	glMatrixMode(GL_PROJECTION);
	gluPerspective(45.0, (float)xres / (float)yres, 1.0, 100.0);
	glMatrixMode(GL_MODELVIEW);

	init_sdr();

	if(!(vs = load_vertex_shader("sdr.vs.glsl"))) {
		fprintf(stderr, "failed to load vertex shader\n");
		return EXIT_FAILURE;
	}

	if(!(pslib = load_pixel_shader("sdr_ref_models.glsl"))) {
		fprintf(stderr, "failed to load reflection models shader\n");
		return EXIT_FAILURE;
	}

	if(!(ps = load_pixel_shader("sdr.ps.glsl"))) {
		fprintf(stderr, "failed to load pixel shader\n");
		return EXIT_FAILURE;
	}

	if(!(prog = create_program())) {
		fprintf(stderr, "failed to create GPU program\n");
		return EXIT_FAILURE;
	}
	attach_shader(prog, vs);
	attach_shader(prog, pslib);
	attach_shader(prog, ps);

	if(link_program(prog) == -1) {
		fprintf(stderr, "failed to link GPU program\n");
		return EXIT_FAILURE;
	}

	glutMainLoop();
	return 0;
}

void setup_lights(void)
{
	float lpos[] = {-6, 8, 10, 1};
	glLightfv(GL_LIGHT0, GL_POSITION, lpos);
}

void redraw(void)
{
	glClearColor(0, 0, 0, 0);
	glClearDepth(1.0);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslatef(-cam_x, -cam_y, -cam_z);
	glRotatef(cam_pitch, 1, 0, 0);
	glRotatef(cam_rot, 0, 1, 0);

	setup_lights();

	glPushMatrix();
	glTranslatef(0, 0.8, 0);

	bind_program(prog);
	set_uniform_float(prog, "rough", roughness);
	set_uniform_float(prog, "specularity", specularity);
	set_uniform_float(prog, "ior", ior);
	set_uniform_float(prog, "dif_sdr", dif_sdr);
	set_uniform_float(prog, "spec_sdr", spec_sdr);
	glFrontFace(GL_CW);
	set_material_color(0.55, 0.20, 0.10, 1.0);
	/*set_material_color(0.87, 0.72, 0.62, 1.0);*/
	glutSolidTeapot(1.0);
	glFrontFace(GL_CCW);
	bind_program(0);

	glPopMatrix();
	

	glutSwapBuffers();
}

void set_material_color(float r, float g, float b, float a)
{
	float col[4];
	col[0] = r;
	col[1] = g;
	col[2] = b;
	col[3] = a;
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, col);
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, spec_pow);
}

int mod_rough, mod_spec, mod_ior, mod_spow;

void key_handler(unsigned char key, int x, int y)
{
	switch(key) {
	case 27:
		exit(0);

	case '1':
		mod_rough = 1;
		break;

	case '2':
		mod_spec = 1;
		break;

	case '3':
		mod_ior = 1;
		break;

	case '4':
		mod_spow = 1;
		break;

	default:
		break;
	}
}

void key_up_handler(unsigned char key, int x, int y)
{
	switch(key) {
	case '1':
		mod_rough = 0;
		break;

	case '2':
		mod_spec = 0;
		break;

	case '3':
		mod_ior = 0;
		break;

	case '4':
		mod_spow = 0;
		break;

	default:
		break;
	}
}

void skey_handler(int key, int x, int y)
{
	switch(key) {
	default:
		break;
	}
}

static int prev_x = -1;
static int prev_y = -1;
static int pbn;

void button_handler(int button, int state, int x, int y)
{
	if(state == GLUT_DOWN) {
		prev_x = x;
		prev_y = y;
		pbn = button;
	} else {
		prev_x = -1;
		prev_y = -1;
	}
}


#define MIN(a, b)		((a) < (b) ? (a) : (b))
#define MAX(a, b)		((a) > (b) ? (a) : (b))

void mouse_handler(int x, int y)
{
	if(pbn == GLUT_LEFT_BUTTON) {
		if(mod_rough || mod_spec || mod_ior || mod_spow) {
			float dx = (float)(x - prev_x) / (float)xres;

			if(mod_rough) {
				roughness = MAX(0.0, MIN(roughness + dx, 1.0));
				printf("roughness: %.3f\n", roughness);
			}
			if(mod_spec) {
				specularity = MAX(0.0, MIN(specularity + dx, 1.0));
				printf("specularity: %.3f\n", specularity);
			}
			if(mod_ior) {
				ior = MAX(ior + dx, 0.0);
				printf("ior: %.3f\n", ior);
			}
			if(mod_spow) {
				spec_pow = MAX(spec_pow + dx * 50.0, 0.01);
				printf("specular power: %.3f\n", spec_pow);
			}
		} else {
			/*
			cam_x += (float)(prev_x - x) * pan_offs;
			cam_y += (float)(y - prev_y) * pan_offs;
			*/
		}

		prev_x = x;
		prev_y = y;
		glutPostRedisplay();

	} else if(pbn == GLUT_MIDDLE_BUTTON) {
		cam_rot += (float)(x - prev_x) / 2.0f;
		cam_pitch += (float)(y - prev_y) / 2.0f;
		if(cam_pitch > 90.0) cam_pitch = 90.0;
		if(cam_pitch < -90.0) cam_pitch = -90.0;
		
		prev_x = x;
		prev_y = y;
		glutPostRedisplay();

	} else if(pbn == GLUT_RIGHT_BUTTON) {
		/*
		cam_z -= (float)(prev_y - y) / 2.0f;
		prev_y = y;
		glutPostRedisplay();
		*/
	}
}

void reshape(int x, int y)
{
	glViewport(0, 0, x, y);
	xres = x;
	yres = y;

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45.0, (float)xres / (float)yres, 1.0, 100.0);
	glMatrixMode(GL_MODELVIEW);
}

#ifdef MOUSEWHEEL
void wheel_handler(int unk, int dir, int x, int y)
{
	if(dir > 0) {
		cam_z -= pan_offs * 5.0;
	} else {
		cam_z += pan_offs * 5.0;
	}
	glutPostRedisplay();
}
#endif

void menu_handler(int val)
{
	if(val < 10) {
		dif_sdr = val;
	} else {
		spec_sdr = val - 10;
	}
	glutPostRedisplay();
}
